@@ -23,6 +23,7 @@
  * Authors:
  *	Kristian Høgsberg (krh@redhat.com)
  *	Adam Jackson (ajax@redhat.com)
+ *	Peter Hutterer (peter.hutterer@redhat.com)
  */
 
 #ifdef HAVE_CONFIG_H
@@ -30,26 +31,26 @@
 #endif
 
 #include <X11/keysym.h>
-#include <X11/XF86keysym.h>
-#include <X11/extensions/XIproto.h>
 
 #include <unistd.h>
+#include <errno.h>
+#include <fcntl.h>
 
-#include <misc.h>
 #include <xf86.h>
-#include <xf86str.h>
-#include <xf86_OSproc.h>
 #include <xf86Xinput.h>
 #include <exevents.h>
-#include <mipointer.h>
 #include <xorgVersion.h>
+#ifdef XKB
+#include <xkbsrv.h>
+#endif
 
 #include "evdev.h"
 
-#include <xf86Module.h>
+#ifdef HAVE_PROPERTIES
+#include <X11/Xatom.h>
+#include <evdev-properties.h>
+#endif
 
-#include <errno.h>
-#include <fcntl.h>
 
 /* 2.4 compatibility */
 #ifndef EVIOCGRAB
@@ -89,16 +90,24 @@
 #define MODEFLAG	8
 #define COMPOSEFLAG	16
 
-
 static const char *evdevDefaults[] = {
-    "XkbRules",     "base",
+    "XkbRules",     "evdev",
     "XkbModel",     "evdev",
     "XkbLayout",    "us",
     NULL
 };
 
 static int EvdevOn(DeviceIntPtr);
-static int EvdevCacheCompare(InputInfoPtr pInfo, Bool compare);
+static int EvdevCacheCompare(InputInfoPtr pInfo, BOOL compare);
+
+#ifdef HAVE_PROPERTIES
+static void EvdevInitProperty(DeviceIntPtr dev);
+static int EvdevSetProperty(DeviceIntPtr dev, Atom atom,
+                            XIPropertyValuePtr val, BOOL checkonly);
+static Atom prop_invert = 0;
+static Atom prop_reopen = 0;
+#endif
+
 
 static void
 SetXkbOption(InputInfoPtr pInfo, char *name, char **option)
@@ -155,10 +164,6 @@ PostKbdEvent(InputInfoPtr pInfo, struct input_event *ev, int value)
 	warned[ev->code] = 1;
     }
 
-    /* The X server can't handle keycodes > 255 anyway, just drop them.  */
-    if (code > 255)
-        return;
-
     xf86PostKeyboardEvent(pInfo->dev, code, value);
 }
 
@@ -217,6 +222,7 @@ EvdevReadInput(InputInfoPtr pInfo)
     int len, value;
     int dx, dy;
     unsigned int abs;
+    unsigned int button;
     EvdevPtr pEvdev = pInfo->private;
 
     dx = 0;
@@ -241,11 +247,16 @@ EvdevReadInput(InputInfoPtr pInfo)
             break;
         }
 
+
         /* Get the signed value, earlier kernels had this as unsigned */
         value = ev.value;
 
         switch (ev.type) {
 	case EV_REL:
+	    /* Handle mouse wheel emulation */
+	    if (EvdevWheelEmuFilterMotion(pInfo, &ev))
+		break;
+
             switch (ev.code) {
             case REL_X:
                 dx += value;
@@ -258,7 +269,7 @@ EvdevReadInput(InputInfoPtr pInfo)
             case REL_WHEEL:
                 if (value > 0)
                     PostButtonClicks(pInfo, wheel_up_button, value);
-                if (value < 0)
+                else if (value < 0)
                     PostButtonClicks(pInfo, wheel_down_button, -value);
                 break;
 
@@ -266,7 +277,7 @@ EvdevReadInput(InputInfoPtr pInfo)
             case REL_HWHEEL:
                 if (value > 0)
                     PostButtonClicks(pInfo, wheel_right_button, value);
-                if (value < 0)
+                else if (value < 0)
                     PostButtonClicks(pInfo, wheel_left_button, -value);
                 break;
             }
@@ -292,53 +303,35 @@ EvdevReadInput(InputInfoPtr pInfo)
 		    break;
 
             switch (ev.code) {
-	    /* swap here, pretend we're an X-conformant device. */
-            case BTN_LEFT:
-                if (!EvdevMBEmuFilterEvent(pInfo, ev.code, value))
-                    xf86PostButtonEvent(pInfo->dev, 0, 1, value, 0, 0);
-                break;
-            case BTN_RIGHT:
-                if (!EvdevMBEmuFilterEvent(pInfo, ev.code, value))
-                    xf86PostButtonEvent(pInfo->dev, 0, 3, value, 0, 0);
-                break;
-            case BTN_MIDDLE:
-                EvdevMBEmuEnable(pInfo, FALSE);
-                xf86PostButtonEvent(pInfo->dev, 0, 2, value, 0, 0);
-                break;
-
-            case BTN_SIDE:
-            case BTN_EXTRA:
-            case BTN_FORWARD:
-            case BTN_BACK:
-            case BTN_TASK:
-                xf86PostButtonEvent(pInfo->dev, 0, ev.code - BTN_LEFT + 5,
-                                    value, 0, 0);
-                break;
-
 	    case BTN_TOUCH:
- 	    case BTN_TOOL_PEN:
- 	    case BTN_TOOL_RUBBER:
- 	    case BTN_TOOL_BRUSH:
- 	    case BTN_TOOL_PENCIL:
- 	    case BTN_TOOL_AIRBRUSH:
- 	    case BTN_TOOL_FINGER:
- 	    case BTN_TOOL_MOUSE:
- 	    case BTN_TOOL_LENS:
+	    case BTN_TOOL_PEN:
+	    case BTN_TOOL_RUBBER:
+	    case BTN_TOOL_BRUSH:
+	    case BTN_TOOL_PENCIL:
+	    case BTN_TOOL_AIRBRUSH:
+	    case BTN_TOOL_FINGER:
+	    case BTN_TOOL_MOUSE:
+	    case BTN_TOOL_LENS:
 		pEvdev->tool = value ? ev.code : 0;
 		break;
 
             default:
-		if (ev.code > BTN_TASK && ev.code < KEY_OK) {
-		    /* Some fancy mice with a lot of buttons generate
-		     * button events between BTN_TASK and BTN_JOYSTICK */
-		    if (ev.code < BTN_JOYSTICK)
-			xf86PostButtonEvent(pInfo->dev, 0,
-			                    ev.code - BTN_LEFT + 5,
-			                    value, 0, 0);
-		    break;
-		}
+		button = EvdevUtilButtonEventToButtonNumber(ev.code);
 
-                PostKbdEvent(pInfo, &ev, value);
+		/* Handle drag lock */
+		if (EvdevDragLockFilterEvent(pInfo, button, value))
+		    break;
+
+		if (EvdevWheelEmuFilterButton(pInfo, button, value))
+		   break;
+
+		if (EvdevMBEmuFilterEvent(pInfo, button, value))
+		   break;
+
+		if (button)
+		    xf86PostButtonEvent(pInfo->dev, 0, button, value, 0, 0);
+		else
+		    PostKbdEvent(pInfo, &ev, value);
 		break;
             }
             break;
@@ -363,8 +356,13 @@ EvdevReadInput(InputInfoPtr pInfo)
 	}
     }
 
-    if (dx != 0 || dy != 0)
+    if (dx != 0 || dy != 0) {
+        if (pEvdev->invert_x)
+            dx *= -1;
+        if (pEvdev->invert_y)
+            dy *= -1;
         xf86PostMotionEvent(pInfo->dev, FALSE, 0, 2, dx, dy);
+    }
 
     /*
      * Some devices only generate valid abs coords when BTN_DIGI is
@@ -376,12 +374,19 @@ EvdevReadInput(InputInfoPtr pInfo)
      * just works.
      */
     if (abs && pEvdev->tool) {
-	xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2,
-			    pEvdev->abs_x, pEvdev->abs_y);
+        int abs_x, abs_y;
+        abs_x = pEvdev->abs_x;
+        abs_y = pEvdev->abs_y;
+        if (pEvdev->invert_x)
+            abs_x = pEvdev->max_x - abs_x;
+        if (pEvdev->invert_y)
+            abs_y = pEvdev->max_y - abs_y;
+
+	xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, abs_x, abs_y);
     }
 }
 
-#define TestBit(bit, array) (array[(bit) / LONG_BITS]) & (1L << ((bit) % LONG_BITS))
+#define TestBit(bit, array) (array[(bit) / LONG_BITS]) & (1 << ((bit) % LONG_BITS))
 
 static void
 EvdevPtrCtrlProc(DeviceIntPtr device, PtrCtrl *ctrl)
@@ -682,7 +687,7 @@ EvdevAddKeyClass(DeviceIntPtr device)
 
     static struct { KeySym keysym; CARD8 mask; } modifiers[] = {
         { XK_Shift_L,		ShiftMask },
-        { XK_Shift_R, 		ShiftMask },
+        { XK_Shift_R,		ShiftMask },
         { XK_Control_L,		ControlMask },
         { XK_Control_R,		ControlMask },
         { XK_Caps_Lock,		LockMask },
@@ -695,11 +700,6 @@ EvdevAddKeyClass(DeviceIntPtr device)
         { XK_Mode_switch,	AltLangMask }
     };
 
-    /* TODO:
-     * Ctrl-Alt-Backspace and other Ctrl-Alt-stuff should work
-     * Get keyboard repeat under control (right now caps lock repeats!)
-     */
-
     pInfo = device->public.devicePrivate;
     pEvdev = pInfo->private;
 
@@ -731,12 +731,12 @@ EvdevAddKeyClass(DeviceIntPtr device)
 #ifdef XKB
     else
     {
+	/* sorry, no rules change allowed for you */
+	xf86ReplaceStrOption(pInfo->options, "xkb_rules", "evdev");
         SetXkbOption(pInfo, "xkb_rules", &pEvdev->xkb_rules);
-	if (!pEvdev->xkb_rules)
-	    SetXkbOption(pInfo, "XkbRules", &pEvdev->xkb_rules);
-	/* sorry, no model change allowed for you */
-	xf86ReplaceStrOption(pInfo->options, "xkb_model", "evdev");
         SetXkbOption(pInfo, "xkb_model", &pEvdev->xkb_model);
+	if (!pEvdev->xkb_model)
+	    SetXkbOption(pInfo, "XkbModel", &pEvdev->xkb_rules);
         SetXkbOption(pInfo, "xkb_layout", &pEvdev->xkb_layout);
 	if (!pEvdev->xkb_layout)
 	    SetXkbOption(pInfo, "XkbLayout", &pEvdev->xkb_layout);
@@ -851,22 +851,71 @@ EvdevAddRelClass(DeviceIntPtr device)
 static int
 EvdevAddButtonClass(DeviceIntPtr device)
 {
-    CARD8 map[32];
     InputInfoPtr pInfo;
-    int i;
+    EvdevPtr pEvdev;
 
     pInfo = device->public.devicePrivate;
+    pEvdev = pInfo->private;
 
     /* FIXME: count number of actual buttons */
-    for (i = 0; i < ArrayLength(map); i++)
-        map[i] = i;
-
-    if (!InitButtonClassDeviceStruct(device, ArrayLength(map), map))
+    if (!InitButtonClassDeviceStruct(device, ArrayLength(pEvdev->btnmap),
+                                     pEvdev->btnmap))
         return !Success;
 
     return Success;
 }
 
+/**
+ * Init the button mapping for the device. By default, this is a 1:1 mapping,
+ * i.e. Button 1 maps to Button 1, Button 2 to 2, etc.
+ *
+ * If a mapping has been specified, the mapping is the default, with the
+ * user-defined ones overwriting the defaults.
+ * i.e. a user-defined mapping of "3 2 1" results in a mapping of 3 2 1 4 5 6 ...
+ *
+ * Invalid button mappings revert to the default.
+ *
+ * Note that index 0 is unused, button 0 does not exist.
+ * This mapping is initialised for all devices, but only applied if the device
+ * has buttons (in EvdevAddButtonClass).
+ */
+static void
+EvdevInitButtonMapping(InputInfoPtr pInfo)
+{
+    int         i, nbuttons     = 1;
+    char       *mapping         = NULL;
+    EvdevPtr    pEvdev          = pInfo->private;
+
+    /* Check for user-defined button mapping */
+    if ((mapping = xf86CheckStrOption(pInfo->options, "ButtonMapping", NULL)))
+    {
+        char    *s  = " ";
+        int     btn = 0;
+
+        xf86Msg(X_CONFIG, "%s: ButtonMapping '%s'\n", pInfo->name, mapping);
+        while (s && *s != '\0' && nbuttons < EVDEV_MAXBUTTONS)
+        {
+            btn = strtol(mapping, &s, 10);
+
+            if (s == mapping || btn < 0 || btn > EVDEV_MAXBUTTONS)
+            {
+                xf86Msg(X_ERROR,
+                        "%s: ... Invalid button mapping. Using defaults\n",
+                        pInfo->name);
+                nbuttons = 1; /* ensure defaults start at 1 */
+                break;
+            }
+
+            pEvdev->btnmap[nbuttons++] = btn;
+            mapping = s;
+        }
+    }
+
+    for (i = nbuttons; i < ArrayLength(pEvdev->btnmap); i++)
+        pEvdev->btnmap[i] = i;
+
+}
+
 static int
 EvdevInit(DeviceIntPtr device)
 {
@@ -896,6 +945,17 @@ EvdevInit(DeviceIntPtr device)
     else if (pEvdev->flags & EVDEV_ABSOLUTE_EVENTS)
         EvdevAddAbsClass(device);
 
+#ifdef HAVE_PROPERTIES
+    /* We drop the return value, the only time we ever want the handlers to
+     * unregister is when the device dies. In which case we don't have to
+     * unregister anyway */
+    XIRegisterPropertyHandler(device, EvdevSetProperty, NULL, NULL);
+    EvdevInitProperty(device);
+    EvdevMBEmuInitProperty(device);
+    EvdevWheelEmuInitProperty(device);
+    EvdevDragLockInitProperty(device);
+#endif
+
     return Success;
 }
 
@@ -936,11 +996,6 @@ EvdevOn(DeviceIntPtr device)
     } else
     {
         xf86AddEnabledDevice(pInfo);
-        if ((pEvdev->flags & EVDEV_BUTTON_EVENTS) &&
-            !(pEvdev->flags & EVDEV_INITIALIZED))
-        {
-            EvdevMBEmuPreInit(pInfo);
-        }
         pEvdev->flags |= EVDEV_INITIALIZED;
         device->public.on = TRUE;
     }
@@ -999,29 +1054,6 @@ EvdevProc(DeviceIntPtr device, int what)
     return Success;
 }
 
-static Bool
-EvdevConvert(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
-	     int v3, int v4, int v5, int *x, int *y)
-{
-    EvdevPtr pEvdev = pInfo->private;
-    int screenWidth = screenInfo.screens[pEvdev->screen]->width;
-    int screenHeight = screenInfo.screens[pEvdev->screen]->height;
-
-    if (first != 0 || num != 2)
-	return FALSE;
-
-    /* on absolute touchpads, don't warp on initial touch */
-    if (pEvdev->flags & EVDEV_TOUCHPAD) {
-	*x = v0;
-	*y = v0;
-	return TRUE;
-    }
-
-    *x = (v0 - pEvdev->min_x) * screenWidth / (pEvdev->max_x - pEvdev->min_x);
-    *y = (v1 - pEvdev->min_y) * screenHeight / (pEvdev->max_y - pEvdev->min_y);
-
-    return TRUE;
-}
 /**
  * Get as much information as we can from the fd and cache it.
  * If compare is True, then the information retrieved will be compared to the
@@ -1031,7 +1063,7 @@ EvdevConvert(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
  * @return Success if the information was cached, or !Success otherwise.
  */
 static int
-EvdevCacheCompare(InputInfoPtr pInfo, Bool compare)
+EvdevCacheCompare(InputInfoPtr pInfo, BOOL compare)
 {
     EvdevPtr pEvdev = pInfo->private;
     int i;
@@ -1131,6 +1163,7 @@ EvdevCacheCompare(InputInfoPtr pInfo, Bool compare)
 
 error:
     return !Success;
+
 }
 
 static int
@@ -1139,7 +1172,7 @@ EvdevProbe(InputInfoPtr pInfo)
     long key_bitmask[NBITS(KEY_MAX)];
     long rel_bitmask[NBITS(REL_MAX)];
     long abs_bitmask[NBITS(ABS_MAX)];
-    int i, has_axes, has_buttons, has_keys;
+    int i, has_axes, has_keys, num_buttons;
     EvdevPtr pEvdev = pInfo->private;
 
     if (ioctl(pInfo->fd, EVIOCGRAB, (void *)1)) {
@@ -1154,7 +1187,7 @@ EvdevProbe(InputInfoPtr pInfo)
         ioctl(pInfo->fd, EVIOCGRAB, (void *)0);
     }
 
-    if (ioctl(pInfo->fd, 
+    if (ioctl(pInfo->fd,
               EVIOCGBIT(EV_REL, sizeof(rel_bitmask)), rel_bitmask) < 0) {
         xf86Msg(X_ERROR, "ioctl EVIOCGBIT failed: %s\n", strerror(errno));
         return 1;
@@ -1173,15 +1206,15 @@ EvdevProbe(InputInfoPtr pInfo)
     }
 
     has_axes = FALSE;
-    has_buttons = FALSE;
     has_keys = FALSE;
+    num_buttons = 0;
 
     if (TestBit(REL_X, rel_bitmask) && TestBit(REL_Y, rel_bitmask)) {
         xf86Msg(X_INFO, "%s: Found x and y relative axes\n", pInfo->name);
 	pEvdev->flags |= EVDEV_RELATIVE_EVENTS;
 	has_axes = TRUE;
     }
-      
+
     if (TestBit(ABS_X, abs_bitmask) && TestBit(ABS_Y, abs_bitmask)) {
         xf86Msg(X_INFO, "%s: Found x and y absolute axes\n", pInfo->name);
 	pEvdev->flags |= EVDEV_ABSOLUTE_EVENTS;
@@ -1193,10 +1226,19 @@ EvdevProbe(InputInfoPtr pInfo)
 	has_axes = TRUE;
     }
 
-    if (TestBit(BTN_LEFT, key_bitmask)) {
-        xf86Msg(X_INFO, "%s: Found mouse buttons\n", pInfo->name);
-	pEvdev->flags |= EVDEV_BUTTON_EVENTS;
-	has_buttons = TRUE;
+    /* count all buttons */
+    for (i = BTN_MISC; i < BTN_JOYSTICK; i++)
+    {
+        if (TestBit(i, key_bitmask))
+            num_buttons++;
+    }
+
+    if (num_buttons)
+    {
+        pEvdev->flags |= EVDEV_BUTTON_EVENTS;
+        pEvdev->buttons = num_buttons;
+        xf86Msg(X_INFO, "%s: Found %d mouse buttons\n", pInfo->name,
+                num_buttons);
     }
 
     for (i = 0; i < BTN_MISC; i++)
@@ -1209,9 +1251,9 @@ EvdevProbe(InputInfoPtr pInfo)
 	has_keys = TRUE;
     }
 
-    if (has_axes && has_buttons) {
+    if (has_axes && num_buttons) {
         xf86Msg(X_INFO, "%s: Configuring as mouse\n", pInfo->name);
-	pInfo->flags |= XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS | 
+	pInfo->flags |= XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS |
 	    XI86_CONFIGURED;
 	pInfo->type_name = XI_MOUSE;
     }
@@ -1236,11 +1278,11 @@ EvdevProbe(InputInfoPtr pInfo)
     return 0;
 }
 
+
 static InputInfoPtr
 EvdevPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 {
     InputInfoPtr pInfo;
-    MessageType deviceFrom = X_CONFIG;
     const char *device;
     EvdevPtr pEvdev;
 
@@ -1257,7 +1299,7 @@ EvdevPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pInfo->control_proc = NULL;
     pInfo->close_proc = NULL;
     pInfo->switch_mode = NULL;
-    pInfo->conversion_proc = EvdevConvert;
+    pInfo->conversion_proc = NULL;
     pInfo->reverse_conversion_proc = NULL;
     pInfo->dev = NULL;
     pInfo->private_flags = 0;
@@ -1270,18 +1312,15 @@ EvdevPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pInfo->private = pEvdev;
 
     xf86CollectInputOptions(pInfo, evdevDefaults, NULL);
-    xf86ProcessCommonOptions(pInfo, pInfo->options); 
+    xf86ProcessCommonOptions(pInfo, pInfo->options);
 
-    pEvdev->screen = xf86SetIntOption(pInfo->options, "ScreenNumber", 0);
     /*
      * We initialize pEvdev->tool to 1 so that device that doesn't use
      * proximity will still report events.
      */
     pEvdev->tool = 1;
 
-    device = xf86CheckStrOption(dev->commonOptions, "Path", NULL);
-    if (!device)
-	device = xf86CheckStrOption(dev->commonOptions, "Device", NULL);
+    device = xf86CheckStrOption(dev->commonOptions, "Device", NULL);
     if (!device) {
         xf86Msg(X_ERROR, "%s: No device specified.\n", pInfo->name);
 	xf86DeleteInput(pInfo, 0);
@@ -1290,7 +1329,7 @@ EvdevPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 
     pEvdev->device = device;
 
-    xf86Msg(deviceFrom, "%s: Device: \"%s\"\n", pInfo->name, device);
+    xf86Msg(X_CONFIG, "%s: Device: \"%s\"\n", pInfo->name, device);
     do {
         pInfo->fd = open(device, O_RDWR, 0);
     } while (pInfo->fd < 0 && errno == EINTR);
@@ -1302,9 +1341,12 @@ EvdevPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     }
 
     pEvdev->reopen_attempts = xf86SetIntOption(pInfo->options, "ReopenAttempts", 10);
+    pEvdev->invert_x = xf86SetBoolOption(pInfo->options, "InvertX", FALSE);
+    pEvdev->invert_y = xf86SetBoolOption(pInfo->options, "InvertY", FALSE);
 
-    pEvdev->noXkb = noXkbExtension;
-    /* parse the XKB options during kbd setup */
+    pEvdev->noXkb = noXkbExtension; /* parse the XKB options during kbd setup */
+
+    EvdevInitButtonMapping(pInfo);
 
     if (EvdevProbe(pInfo)) {
 	close(pInfo->fd);
@@ -1314,6 +1356,13 @@ EvdevPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 
     EvdevCacheCompare(pInfo, FALSE); /* cache device data */
 
+    if (pEvdev->flags & EVDEV_BUTTON_EVENTS)
+    {
+        EvdevMBEmuPreInit(pInfo);
+        EvdevWheelEmuPreInit(pInfo);
+        EvdevDragLockPreInit(pInfo);
+    }
+
     return pInfo;
 }
 
@@ -1362,3 +1411,117 @@ _X_EXPORT XF86ModuleData evdevModuleData =
     EvdevPlug,
     EvdevUnplug
 };
+
+
+/* Return an index value for a given button event code
+ * returns 0 on non-button event.
+ */
+unsigned int
+EvdevUtilButtonEventToButtonNumber(int code)
+{
+    unsigned int button = 0;
+
+    switch(code) {
+    case BTN_LEFT:
+	button = 1;
+	break;
+
+    case BTN_RIGHT:
+	button = 3;
+	break;
+
+    case BTN_MIDDLE:
+	button = 2;
+	break;
+
+    case BTN_SIDE:
+    case BTN_EXTRA:
+    case BTN_FORWARD:
+    case BTN_BACK:
+    case BTN_TASK:
+	button = (code - BTN_LEFT + 5);
+	break;
+
+    default:
+	if ((code > BTN_TASK) && (code < KEY_OK)) {
+	    if (code < BTN_JOYSTICK)
+		button = (code - BTN_LEFT + 5);
+	}
+    }
+
+    if (button > EVDEV_MAXBUTTONS)
+	return 0;
+
+    return button;
+}
+
+#ifdef HAVE_PROPERTIES
+static void
+EvdevInitProperty(DeviceIntPtr dev)
+{
+    InputInfoPtr pInfo  = dev->public.devicePrivate;
+    EvdevPtr     pEvdev = pInfo->private;
+    int          rc;
+    BOOL         invert[2];
+    char         reopen;
+
+
+    if (pEvdev->flags & (EVDEV_RELATIVE_EVENTS | EVDEV_ABSOLUTE_EVENTS))
+    {
+        invert[0] = pEvdev->invert_x;
+        invert[1] = pEvdev->invert_y;
+
+        prop_invert = MakeAtom(EVDEV_PROP_INVERT_AXES, strlen(EVDEV_PROP_INVERT_AXES), TRUE);
+
+        rc = XIChangeDeviceProperty(dev, prop_invert, XA_INTEGER, 8,
+                PropModeReplace, 2,
+                invert, FALSE);
+        if (rc != Success)
+            return;
+
+        XISetDevicePropertyDeletable(dev, prop_invert, FALSE);
+    }
+
+    prop_reopen = MakeAtom(EVDEV_PROP_REOPEN, strlen(EVDEV_PROP_REOPEN),
+            TRUE);
+
+    reopen = pEvdev->reopen_attempts;
+    rc = XIChangeDeviceProperty(dev, prop_reopen, XA_INTEGER, 8,
+                                PropModeReplace, 1, &reopen, FALSE);
+    if (rc != Success)
+        return;
+
+    XISetDevicePropertyDeletable(dev, prop_reopen, FALSE);
+}
+
+static int
+EvdevSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
+                 BOOL checkonly)
+{
+    InputInfoPtr pInfo  = dev->public.devicePrivate;
+    EvdevPtr     pEvdev = pInfo->private;
+
+    if (atom == prop_invert)
+    {
+        BOOL* data;
+        if (val->format != 8 || val->size != 2 || val->type != XA_INTEGER)
+            return BadMatch;
+
+        if (!checkonly)
+        {
+            data = (BOOL*)val->data;
+            pEvdev->invert_x = data[0];
+            pEvdev->invert_y = data[1];
+        }
+    } else if (atom == prop_reopen)
+    {
+        if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
+            return BadMatch;
+
+        if (!checkonly)
+            pEvdev->reopen_attempts = *((CARD8*)val->data);
+    }
+
+    return Success;
+}
+#endif
